#ifndef __LINUX_SPINLOCK_H
#define __LINUX_SPINLOCK_H

/*Solaris Include*/
#include <sys/ksynch.h>

/*this is stuff added to support spin-locks using mutexes.*/
#define NORMAL_LOCK  0
#define IRQ_LOCK     1
#define IRQSAVE_LOCK 2
#define BH_LOCK      3

typedef struct pits_spinlock
{
  int initialized;
  int initialized_iblock;
  int locked;
  kmutex_t *normal_lock;
  kcondvar_t *normal_cv;
  kmutex_t *iblock_lock;
  int critical;
  struct pits_spinlock *next;
}spinlock_t;

void pits_add_spin_lock(spinlock_t *lock); /*defined in pits_spinlock.c*/
int pits_spin_lock_init_iblock(spinlock_t *lock);
void pits_fini_spin_lock(spinlock_t *lock);

void pits_spin_lock(spinlock_t *lock,int type);
void pits_spin_lock_normal(spinlock_t *lock);
void pits_spin_lock_irq(spinlock_t *lock);

void pits_spin_unlock(spinlock_t *lock,int type);
void pits_spin_unlock_normal(spinlock_t *lock);
void pits_spin_unlock_irq(spinlock_t *lock);

int pits_spin_trylock(spinlock_t *lock,int type);




/*
 * These are the generic versions of the spinlocks and read-write
 * locks..
 */

#ifndef local_irq_save
#define local_irq_save(flags) flags=0
#endif


//#define spin_lock(lock)         (void)(lock) /* Not "unused variable". */
#define spin_lock(lock) pits_spin_lock(lock,NORMAL_LOCK)
#define spin_trylock(lock) pits_spin_trylock(lock,NORMAL_LOCK)


#define spin_lock_irqsave(lock,flags) flags = 0;pits_spin_lock(lock,IRQSAVE_LOCK)
#define spin_lock_irq(lock) pits_spin_lock(lock,IRQ_LOCK);

#define spin_unlock_irqrestore(lock, flags)	flags=0;pits_spin_unlock(lock,IRQSAVE_LOCK)
#define spin_unlock_irq(lock)			pits_spin_unlock(lock,IRQ_LOCK);
#define spin_unlock_bh(lock)			pits_spin_unlock(lock,BH_LOCK);
#define spin_unlock(lock)                       pits_spin_unlock(lock,NORMAL_LOCK);


typedef struct pits_rwlock
{
  int initialized;
  int write_lock;
  int readers;
  kmutex_t *normal_lock;
  kcondvar_t *normal_cv;
  kmutex_t *iblock_lock;
  int critical;
  struct pits_rwlock *next;
}rwlock_t;

void pits_add_rw_lock(rwlock_t *lock); /*defined in pits_spinlock.c*/
void pits_read_lock  (rwlock_t *lock,int type);
void pits_write_lock (rwlock_t *lock,int type);
void pits_read_unlock(rwlock_t *lock,int type);
void pits_write_unlock(rwlock_t *lock,int type);


#define read_lock_irqsave(lock, flags)		flags=0; pits_read_lock(lock,IRQSAVE_LOCK)
#define read_lock_irq(lock)			pits_read_lock(lock,IRQ_LOCK)
#define read_lock_bh(lock)			pits_read_lock(lock,BH_LOCK)

#define write_lock_irqsave(lock, flags)		flags=0;pits_write_lock(lock,IRQSAVE_LOCK)
#define write_lock_irq(lock)			pits_write_lock(lock,IRQ_LOCK)
#define write_lock_bh(lock)			pits_write_lock(lock,BH_LOCK)

#define write_unlock_irqrestore(lock, flags)	pits_write_unlock(lock,IRQSAVE_LOCK)
#define write_unlock_irq(lock)			pits_write_unlock(lock,IRQ_LOCK)
#define write_unlock_bh(lock)			pits_write_unlock(lock,BH_LOCK)

#define read_unlock_irqrestore(lock, flags)	pits_read_unlock(lock,IRQSAVE_LOCK)
#define read_unlock_irq(lock)			pits_read_unlock(lock,IRQ_LOCK)
#define read_unlock_bh(lock)			pits_read_unlock(lock,BH_LOCK)



/*
 * Your basic spinlocks, allowing only a single CPU anywhere
 *
 * Most gcc versions have a nasty bug with empty initializers.
 */

#define SPIN_LOCK_UNLOCKED (spinlock_t) { initialized:0,locked:0,next:NULL}
#define SPIN_LOCK_LOCKED(spinlock_t) {initialized:0,locked:1,next:NULL}

#define spin_lock_init(lock) pits_add_spin_lock(lock)

/*
 * Read-write spinlocks, allowing multiple readers
 * but only one writer.
 *
 * NOTE! it is quite common to have readers in interrupts
 * but no interrupt writers. For those circumstances we
 * can "mix" irq-safe locks - any writer needs to get a
 * irq-safe write-lock, but readers can get non-irqsafe
 * read-locks.
 *
 * Most gcc versions have a nasty bug with empty initializers.
 */
#define RW_LOCK_UNLOCKED (rwlock_t) {initialized:0,write_lock:0,next:NULL}
#define RW_LOCK_LOCKED (rwlock_t) {initialized:0,write_lock:1,next:NULL}

#define rwlock_init(lock)	pits_add_rw_lock(lock)


#define read_lock(lock)	        pits_read_lock(lock,NORMAL_LOCK)
#define read_unlock(lock)	pits_read_unlock(lock,NORMAL_LOCK)
#define write_lock(lock)	pits_write_lock(lock,NORMAL_LOCK)
#define write_unlock(lock)	pits_write_unlock(lock,NORMAL_LOCK)

/* "lock on reference count zero" */
#ifndef ATOMIC_DEC_AND_LOCK
#include <asm/atomic.h>
extern int atomic_dec_and_lock(atomic_t *atomic, spinlock_t *lock);
#endif



#endif
