/*
 *  linux/kernel/timer.c
 *
 *  Kernel internal timers, kernel timekeeping, basic process system calls
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 *
 *  1997-01-28  Modified by Finn Arne Gangstad to make timers scale better.
 *
 *  1997-09-10  Updated NTP code according to technical memorandum Jan '96
 *              "A Kernel Model for Precision Timekeeping" by Dave Mills
 *  1998-12-24  Fixed a xtime SMP race (we need the xtime_lock rw spinlock to
 *              serialize accesses to xtime/lost_ticks).
 *                              Copyright (C) 1998  Andrea Arcangeli
 *  1999-03-10  Improved NTP compatibility by Ulrich Windl
 */

#include <linux/config.h>
#include <linux/mm.h>
//#include <linux/timex.h>
#include <linux/delay.h>
#include <linux/smp_lock.h>
#include <linux/interrupt.h>
//#include <linux/kernel_stat.h>
#include <linux/tqueue.h>
#include <asm/uaccess.h>

//#include <linux/tqueue.h>

/*
 * Timekeeping variables
 */

long tick = (1000000 + HZ/2) / HZ;	/* timer interrupt period */

/* The current time */
struct timeval xtime __attribute__ ((aligned (16)));

/* Don't completely fail for HZ > 500.  */
int tickadj = 500/HZ ? : 1;		/* microsecs */

//DECLARE_TASK_QUEUE(tq_timer);
//DECLARE_TASK_QUEUE(tq_immediate);

/*
 * phase-lock loop variables
 */
/* TIME_ERROR prevents overwriting the CMOS clock */
//int time_state = TIME_OK;		/* clock synchronization status	*/
//int time_status = STA_UNSYNC;		/* clock status bits		*/
//long time_offset;			/* time adjustment (us)		*/
//long time_constant = 2;			/* pll time constant		*/
//long time_tolerance = MAXFREQ;		/* frequency tolerance (ppm)	*/
//long time_precision = 1;		/* clock precision (us)		*/
//long time_maxerror = NTP_PHASE_LIMIT;	/* maximum error (us)		*/
//long time_esterror = NTP_PHASE_LIMIT;	/* estimated error (us)		*/
//long time_phase;			/* phase offset (scaled us)	*/
//long time_freq = ((1000000 + HZ/2) % HZ - HZ/2) << SHIFT_USEC;
					/* frequency offset (scaled ppm)*/
//long time_adj;				/* tick adjust (scaled 1 / HZ)	*/
//long time_reftime;			/* time at last adjustment (s)	*/

//long time_adjust;
//long time_adjust_step;

//unsigned long event;

//extern int do_setitimer(int, struct itimerval *, struct itimerval *);

//unsigned long volatile jiffies;

unsigned int * prof_buffer;
unsigned long prof_len;
unsigned long prof_shift;

/*
 * Event timer code
 */
#define TVN_BITS 6
#define TVR_BITS 8
#define TVN_SIZE (1 << TVN_BITS)
#define TVR_SIZE (1 << TVR_BITS)
#define TVN_MASK (TVN_SIZE - 1)
#define TVR_MASK (TVR_SIZE - 1)

struct timer_vec {
	int index;
	struct list_head vec[TVN_SIZE];
};

struct timer_vec_root {
	int index;
	struct list_head vec[TVR_SIZE];
};

static struct timer_vec tv5;
static struct timer_vec tv4;
static struct timer_vec tv3;
static struct timer_vec tv2;
static struct timer_vec_root tv1;

static struct timer_vec * const tvecs[] = {
	(struct timer_vec *)&tv1, &tv2, &tv3, &tv4, &tv5
};

#define NOOF_TVECS (sizeof(tvecs) / sizeof(tvecs[0]))

void init_timervecs (void)
{
	int i;

	for (i = 0; i < TVN_SIZE; i++) {
		INIT_LIST_HEAD(tv5.vec + i);
		INIT_LIST_HEAD(tv4.vec + i);
		INIT_LIST_HEAD(tv3.vec + i);
		INIT_LIST_HEAD(tv2.vec + i);
	}
	for (i = 0; i < TVR_SIZE; i++)
		INIT_LIST_HEAD(tv1.vec + i);
}

static unsigned long timer_jiffies;

static inline void internal_add_timer(struct timer_list *timer)
{
	/*
	 * must be cli-ed when calling this
	 */
	unsigned long expires = timer->expires;
	unsigned long idx = expires - timer_jiffies;
	struct list_head * vec;

	if (idx < TVR_SIZE) {
		int i = expires & TVR_MASK;
		vec = tv1.vec + i;
	} else if (idx < 1 << (TVR_BITS + TVN_BITS)) {
		int i = (expires >> TVR_BITS) & TVN_MASK;
		vec = tv2.vec + i;
	} else if (idx < 1 << (TVR_BITS + 2 * TVN_BITS)) {
		int i = (expires >> (TVR_BITS + TVN_BITS)) & TVN_MASK;
		vec =  tv3.vec + i;
	} else if (idx < 1 << (TVR_BITS + 3 * TVN_BITS)) {
		int i = (expires >> (TVR_BITS + 2 * TVN_BITS)) & TVN_MASK;
		vec = tv4.vec + i;
	} else if ((signed long) idx < 0) {
		/* can happen if you add a timer with expires == jiffies,
		 * or you set a timer to go off in the past
		 */
		vec = tv1.vec + tv1.index;
	} else if (idx <= 0xffffffffUL) {
		int i = (expires >> (TVR_BITS + 3 * TVN_BITS)) & TVN_MASK;
		vec = tv5.vec + i;
	} else {
		/* Can only get here on architectures with 64-bit jiffies */
		INIT_LIST_HEAD(&timer->list);
		return;
	}
	/*
	 * Timers are FIFO!
	 */
	list_add(&timer->list, vec->prev);
}

/* Initialize both explicitly - let's try to have them in the same cache line */
spinlock_t timerlist_lock = SPIN_LOCK_UNLOCKED;

#ifdef CONFIG_SMP
volatile struct timer_list * volatile running_timer;
#define timer_enter(t) do { running_timer = t; mb(); } while (0)
#define timer_exit() do { running_timer = NULL; } while (0)
#define timer_is_running(t) (running_timer == t)
#define timer_synchronize(t) while (timer_is_running(t)) barrier()
#else
#define timer_enter(t)		do { } while (0)
#define timer_exit()		do { } while (0)
#endif




void pits_timeout(void *t_list)
{
  struct timer_list *timer = (struct timer_list *)t_list;

  DEBUGCMD(printk("inside pits_timeout\n"));

  mutex_enter(&(timer -> timer_mutex));

  if (timer -> active == TRUE)
    {
      timer -> active = FALSE;
      timer -> fired  = TRUE;
      mutex_exit(&(timer -> timer_mutex));
      printk("fire timer\n");
      timer -> function(timer -> data);
    }
  else
    {
      mutex_exit(&(timer -> timer_mutex));
      printk("warning timer deactivated\n");
    }
}



void add_timer(struct timer_list * timer)
{
  clock_t offset = ddi_get_lbolt();
  clock_t ticks = timer -> expires - offset;
  
  mutex_enter(&(timer -> timer_mutex));
  if (timer -> active == TRUE)
    {
      printk("warning timer already added...");
      mutex_exit(&(timer -> timer_mutex));
      return;
    }
  mutex_exit(&(timer -> timer_mutex));

  if (offset > timer -> expires) 
    {
      printk("warning negative time expires %lu current %lu",
	     timer -> expires,
	     offset);
      ticks = 1;
    }
  mutex_enter(&(timer -> timer_mutex));
  timer -> active = TRUE;
  timer -> fired = FALSE;
  //mutex_exit(&(timer -> timer_mutex));
  
  //printk("adding timer to timeout in %lu ticks",ticks);
  //printk("thats %lu seconds",drv_hztousec(ticks)/1000000);
  timer -> timer_id = timeout(pits_timeout,timer,ticks);
  mutex_exit(&(timer -> timer_mutex));
}


/*****************************************/
/*So deltimer returns zero if the removal*/
/* of a timer is unsucessful, i.e., the  */
/*timer has fired.                       */
/*****************************************/
int del_timer(struct timer_list * timer)
{
  clock_t cret;
  //DEBUGCMD(printk("inside delete timer"));

  mutex_enter(&(timer -> timer_mutex));
  if (timer -> active == 0)
    {
      mutex_exit(&(timer -> timer_mutex));
      return 0;
    }

  timer -> active = 0;
  mutex_exit(&(timer -> timer_mutex));
  
  cret = untimeout(timer -> timer_id);
  
  if (cret < 0) 
    {
      printk("warning timer not found\n");
    }
  return cret != 0;
}

int mod_timer(struct timer_list *timer, unsigned long expires)
{

  //printk("inside mod_timer");

  mutex_enter(&(timer -> timer_mutex));
  if (timer -> active == 1)
    {
      mutex_exit(&(timer -> timer_mutex));
      del_timer(timer);
    }
  timer -> expires = expires;
  
  add_timer(timer);
  
  return 0;
}

void tqueue_bh(void)
{
	run_task_queue(&tq_timer);
}

void immediate_bh(void)
{
	run_task_queue(&tq_immediate);
}
