/**********************************************/
/*PITS project                                */
/*pits_kernel/pits_bh.c                       */
/*bottom half handler for pits drivers.       */
/**********************************************/
/*taken from /linux/softirq.c*/

#include <linux/config.h>
#include <linux/mm.h>
//#include <linux/kernel_stat.h>
#include <linux/interrupt.h>
#include <linux/smp_lock.h>
#include <linux/init.h>
#include <linux/tqueue.h>

/***************/
/*PITS INCLUDES*/
/***************/
#include <misc/pits_bh.h>


void tqueue_bh(void);

/* Old style BHs */
static void (*bh_base[32])(void);
static struct tasklet_struct bh_task_vec[32];

/* BHs are serialized by spinlock global_bh_lock.

   It is still possible to make synchronize_bh() as
   spin_unlock_wait(&global_bh_lock). This operation is not used
   by kernel now, so that this lock is not made private only
   due to wait_on_irq().

   It can be removed only after auditing all the BHs.
 */
spinlock_t global_bh_lock = SPIN_LOCK_UNLOCKED;
spinlock_t tqueue_lock = SPIN_LOCK_UNLOCKED;

static void bh_action(unsigned long nr)
{
  
  int cpu = smp_processor_id();

  //printk("inside bh_action...\n");

  if (!spin_trylock(&global_bh_lock))
    goto resched;
  
  if (!hardirq_trylock(cpu))
    goto resched_unlock;
  
  if (nr < 0 || nr > 31)
    {
      printk("nr is bad! %d\n",nr);
    }
  if (bh_base[nr])
    {
      bh_base[nr]();
    }
  else
    {
      printk("warning, NULL bh %d\n",(int)nr);
    }
  hardirq_endlock(cpu);
  spin_unlock(&global_bh_lock);
  return;

 resched_unlock:
  printk("bh() : unlock\n");
  spin_unlock(&global_bh_lock);
 resched:
  printk("bh() : rescheduling\n");
  mark_bh(nr);
  
}
/**********************************/
/*init_bh() : initializes the bh  */
/*with a routine.                 */
/**********************************/
void init_bh(int nr, void (*routine)(void))
{
  bh_base[nr] = routine;
  mb();
}
/***********************************/
/*remove_bh() : removes a bh from  */
/*the bh vector.                   */
/***********************************/
void remove_bh(int nr)
{
  //	tasklet_kill(bh_task_vec+nr);
	bh_base[nr] = NULL;
}

void pits_tqueue_timeout(void *data)
{
  int timer_id;
  mark_bh(TQUEUE_BH);
  timer_id = timeout(pits_tqueue_timeout,0,10);
}
/**************************************************/
//void __init softirq_init()
/*init_bh_vec() : initializes the bh_vector       */
/*just calls tasklet_init a bunch of times..      */
/**************************************************/
int init_bh_vec(dev_info_t *dip)
{
	int i;
	int timer_id;
	for (i=0; i<32; i++)
	  {
	    tasklet_init(bh_task_vec+i, bh_action, i);

	    init_bh(i,NULL);
	  }

	init_bh_tq_immediate();
	/*
	open_softirq(TASKLET_SOFTIRQ, tasklet_action, NULL);
	open_softirq(HI_SOFTIRQ, tasklet_hi_action, NULL);
	*/
	init_bh(TQUEUE_BH,tqueue_bh);

	//timer_id = timeout(pits_tqueue_timeout,0,10);


	//set_initialized(dip,BH_INIT);

	return 0;
}

/***********************************/
/*mark_bh() : readies the bh for   */
/*execution.                       */
/***********************************/
void mark_bh(int nr)
{

  //DEBUGCMD(printk("inside mark_bh\n"));

  if (bh_base[nr] == NULL)
    {
      printk("mark_bh: warning null bh! %d\n",nr);
      return;
    }
  DEBUGCMD(
  switch(nr)
    {
    case SCSI_BH:
      //printk("marking SCSI_BH...\n");
      break;
    case IMMEDIATE_BH:
      //printk("marking IMMEDIATE_BH...\n");
      break;
    case TQUEUE_BH:
      //printk("marking TQUEUE_BH...\n");
      break;
    default:
      ;
    }
  );
  tasklet_hi_schedule(bh_task_vec+nr);
  //DEBUGCMD(printk("done mark_bh\n"));
}


/***************************************/
/*fini_bh_vec() :                      */
/*uninitializes the bh_vectors.        */
/***************************************/
int fini_bh_vec(dev_info_t *dip)
{
  //unset_initialized(dip,BH_INIT);
  return DDI_SUCCESS;
}
/******************************************/
/*bh_immediate() : runs everything in the */
/*task_queue tq_immediate                 */
/******************************************/
void bh_immediate(void)
{
  //printk("Inside bh_immediate\n");
  run_task_queue(&tq_immediate);
  //printk("Done bh_immediate\n");
}

/*******************************************/
/*init_bh_tq_immediate() : initializes the */
/*bh value with the function               */
/*******************************************/
void init_bh_tq_immediate()
{
  //printk("init_bh_tq_immediate %d\n",IMMEDIATE_BH);
  init_bh(IMMEDIATE_BH,bh_immediate);
  //printk("done init_bh_tq_immediate\n");
}



int queue_task(struct tq_struct *bh_pointer, task_queue *bh_list)
{
	int ret = 0;
	if (!test_and_set_bit(0,&bh_pointer->sync)) 
	  {
	    unsigned long flags;
	    spin_lock_irqsave(&tqueue_lock, flags);
	    list_add_tail(&bh_pointer->list, bh_list);
	    spin_unlock_irqrestore(&tqueue_lock, flags);
	    ret = 1;

	    //added this so that everytime a task is queued on
	    //the timer, we mark it.
	    if (bh_list == &tq_timer)
	      {
		mark_bh(TQUEUE_BH);
	      }
	  }

	

	return ret;
}


/**********************************************/
/*__run_task_queue() : runs the tasks within  */
/*the task queue.  Uses a global tqueue_lock  */
/**********************************************/
void __run_task_queue(task_queue *list)
{
	struct list_head head, *next;
	unsigned long flags;

	//DEBUGCMD(printk("inside __run_task_queue()\n"));
	spin_lock_irqsave(&tqueue_lock, flags);
	list_add(&head, list);
	list_del_init(list);
	spin_unlock_irqrestore(&tqueue_lock, flags);

	next = head.next;
	while (next != &head) {
		void (*f) (void *);
		struct tq_struct *p;
		void *data;

		p = list_entry(next, struct tq_struct, list);
		next = next->next;
		f = p->routine;
		data = p->data;
		wmb();
		p->sync = 0;
		if (f)
		  {
		    //printk("running task..\n");
		    f(data);
		    //printk("done running task...\n");
		  }
	}
	//printk("__run_task_queue() : done.\n");
}


void tqueue_bh(void)
{
	run_task_queue(&tq_timer);
}
