/*********************************************/
/*slab.c                                     */
/*this file is an attempt to provide support */
/*for the linux calls kmalloc and kfree      */
/*since the sun kernel calls for memory alloc*/
/*requires the size to be provided for       */
/*kmem_free, a linked list is built in to    */
/*manage memory (the static variable klist   */
/*maintains a private variable programs that */
/*use this library (link the .o file)        */
/*Sean McIlwain                              */
/*Barry Henrichs                             */
/*2003                                       */
/*********************************************/

/*Linux includes*/
#include <linux/slab.h>

/*PITS includes*/
//#include <misc/sunlinux.h> /* other function calls */
#include <misc/slab.h>

/*Solaris Includes*/
#include <sys/types.h>    /*added for readability  */
#include <sys/kmem.h>     /*added for readability  */

/*PITS Kernel Includes*/
#include "pits_kernel_state.h"

/* 
linux/slab.h 
extern void kfree(const void *);  
extern void *kmalloc(size_t, int);  
*/ 

//static global variable.
static pits_klist_t mem_list;

int fini_dma_handle(ddi_dma_handle_t *);
void printAlloc();

static void memlist_debug(void *arg)
{
  uint64_t total_allocations;
  uint64_t total_allocated;
  uint64_t total_dma_allocated;
  uint64_t dma_handles;
  mutex_enter(&(mem_list.mutex));
  if (mem_list.timer_running == TRUE)
    {
      total_allocations = mem_list.total_allocations;
      total_allocated = mem_list.total_allocated;
      total_dma_allocated = mem_list.total_dma_allocated;
      dma_handles = mem_list.dma_handles;
      mutex_exit(&(mem_list.mutex));
      printk("memlist debug:\n");
      printk("Total Allocations %lu:\n",total_allocations);
      printk("Total Allocated: %lu\n",total_allocated);
      printk("Total DMA: %lu\n",total_dma_allocated);
      printk("DMA Handles: %lu\n",dma_handles);
      mem_list.timer_id = timeout(memlist_debug,NULL,drv_usectohz(30 * 1000000));
    }
  else
    {
      mutex_exit(&(mem_list.mutex));
    }
}

ddi_dma_handle_t *alloc_dma_handle(int kmflags)
{
  //ddi_dma_handle_t *temp = kmem_cache_alloc(mem_list.dma_handle_cache,
  //kmflags);

  ddi_dma_handle_t *temp = kmem_zalloc(sizeof(ddi_dma_handle_t),kmflags);

  if (temp == NULL)
    {
      printk("alloc_dma_handle(): failed to allocate!\n");
      return NULL;
    }
  if (init_dma_handle(temp) != DDI_SUCCESS)
    {
      printk("alloc_dma_handle(): failed to init!\n");
      kmem_free(temp,sizeof(ddi_dma_handle_t));
      return NULL;
    }
  else
    {
      
      mutex_enter(&(mem_list.mutex));
      mem_list.dma_handles++;
      mutex_exit(&(mem_list.mutex));
    }
  return temp;
}
void release_dma_handle(ddi_dma_handle_t *handle)
{
  //kmem_cache_free(mem_list.dma_handle_cache,handle);
  
  fini_dma_handle(handle);
  kmem_free(handle,sizeof(ddi_dma_handle_t));

  mutex_enter(&(mem_list.mutex));
  mem_list.dma_handles--;
  mutex_exit(&(mem_list.mutex));
}

static int dma_handle_create(void *handle, 
			     void *memlist, 
			     int kmflags)
{
  int ret = 0;
  
  //pits_klist_t *mlist = (pits_klist_t*)memlist;

  //printk("inside pits_kernel_dma_handle_create\n");
  ret = init_dma_handle((ddi_dma_handle_t *)handle);
  
  if (ret != DDI_SUCCESS)
    {
      printk("contructor, error allocating handle!\n");
      ret = DDI_FAILURE;
    }
  else
    {
      ret = 0;
    }
  //printk("done pits_kernel_dma_handle_create: - %d\n",ret);
  return ret;
}



static void dma_handle_destroy(void *handle,void *memlist)
{
  //printk("inside pits_kernel_dma_handle_destroy..\n");
  
  fini_dma_handle((ddi_dma_handle_t*)handle);

  //printk("done pits_kernel_dma_handle_destroy....\n");
  return;
}
 


/********************************************/
/*init_pits_slab()                          */
/*This function will initialize the memory  */
/*allocation structure for the pits driver  */
/********************************************/
int init_pits_slab(dev_info_t *dip)
{
  ddi_dma_attr_t dma_attr = PITS_KERNEL_DMA_ATTR;
  //initialize mem_list structure.
  bzero(&mem_list,sizeof(pits_klist_t));
  mutex_init(&(mem_list.mutex),NULL,MUTEX_DRIVER,NULL);
  mem_list.dip = dip;
  bcopy(&dma_attr,&(mem_list.dma_attr),sizeof(ddi_dma_attr_t));
  mem_list.total_allocations = 0;
  mem_list.total_allocated   = 0;
  mem_list.total_dma_allocated = 0;
  mem_list.dma_handles = 0;
  mem_list.head = NULL;

  mem_list.dma_handle_cache = 
    kmem_cache_create("pdmahandlec",
		      sizeof(ddi_dma_handle_t),      /*bufsize*/
		      0,                             /*align*/
		      dma_handle_create, /*constructor*/
		      dma_handle_destroy,/*destructor*/
		      NULL,                          /*reclaim*/
		      NULL,                            /*argument*/
		      NULL,                          /*vmp*/
		      0);                            /*cflags*/

  if (mem_list.dma_handle_cache == NULL)
    {
      printk("error allocating cache!\n");
      return DDI_FAILURE;
    }
  
  mem_list.timer_running = TRUE;
  //mem_list.timer_id = timeout(memlist_debug,NULL,drv_usectohz(30 * 1000000));
  
  return DDI_SUCCESS;
}

int fini_pits_slab(dev_info_t *dip)
{

  klist_empty();
  mutex_enter(&(mem_list.mutex));
  mem_list.timer_running = FALSE;
  mutex_exit(&(mem_list.mutex));
  //untimeout(mem_list.timer_id);
  
  kmem_cache_destroy(mem_list.dma_handle_cache);

  
  mutex_destroy(&(mem_list.mutex));

  return DDI_SUCCESS;
}

int klist_isempty()
{
  int ans;

  mutex_enter(&(mem_list.mutex));
  if (mem_list.head == NULL)
    {
      ans = TRUE;
    }
  else
    {
      ans = FALSE;
    }
  mutex_exit(&(mem_list.mutex));
  return (ans);
}

/*****************************************/
/*klist_add() :                          */
/*allocates memory with flag, if         */  
/*if successful, adds it to the memory   */
/*list and returns the void pointer to   */
/*the memory.                            */
/*****************************************/
void *klist_add(size_t size,int flag)
{
  kmemptr_t *newnode = NULL;
  //int wait_flag;
  if (flag & GFP_DMA)
    {
      //printk("klist_add() : DMA-safe memory is requested\n");
      newnode = kmalloc_dma(size,flag);
    }
  else
    {
      newnode = kmalloc_normal(size,flag);
    }
  
  if (newnode == NULL) 
    {
      printk("klist_add() : warning cannot allocate kmemptr!\n");
      printAlloc();
      return NULL;
    }
  
  newnode -> size = size;
  newnode -> next = NULL;

  //grab the mutex, since we are modifying the memory list.
  mutex_enter(&(mem_list.mutex));
  if (mem_list.head == NULL)
    {
      mem_list.head = newnode;
    }
  else
    {
      newnode -> next = mem_list.head;
      mem_list.head = newnode;
    }
  mem_list.total_allocations++;
  mem_list.total_allocated += size;
  if (newnode -> flags & GFP_DMA)
    {
      mem_list.total_dma_allocated += size;
    }

  mutex_exit(&(mem_list.mutex));
  //return allocated pointer.
  return (newnode -> allocated);
}

/*************************************************/
/*klist_remove() : This function finds the       */
/*corresponding kmemptr element in mem_list,     */
/*removes it, and releases the memory            */
/*************************************************/
int klist_remove(void *ptr)
{
  kmemptr_t *nextptr = NULL;
  kmemptr_t *prevptr = NULL;

  //grab the mutex.
  mutex_enter(&(mem_list.mutex));

  //scan through the list, trying to find the allocated pointer.
  nextptr = mem_list.head;

  while (nextptr != NULL && nextptr -> allocated != ptr)
    {
      prevptr = nextptr;
      nextptr = nextptr -> next;
    }

  /*didn't find the memory*/
  if (nextptr == NULL)
    {
      mutex_exit(&(mem_list.mutex));
      printk("couldn't find the memory in the list!\n");
      
      return DDI_FAILURE;
    }
  
  //remove the memory node from the list.
  if (prevptr == NULL)
    {
      //remove from the head of the list.
      mem_list.head = mem_list.head -> next;
    }
  else
    {
      //remove from the middle or end of the list.
      prevptr -> next = nextptr -> next;
    }
  mem_list.total_allocations--;
  mem_list.total_allocated = mem_list.total_allocated - nextptr -> size;
  if (nextptr -> flags & GFP_DMA)
    mem_list.total_dma_allocated -= nextptr -> size;

  mutex_exit(&(mem_list.mutex));

  /*free the allocated memory*/
  if (nextptr -> flags & GFP_DMA)
    {
      kfree_dma(nextptr);
    }
  else
    {
      kfree_normal(nextptr);
    }
  return (DDI_SUCCESS);
}
/**********************/ 
/*klist_empty():      */
/*empties the mem_list*/
/**********************/
int klist_empty()
{
  int error;

  while (klist_isempty() == FALSE)
    {
      error = klist_remove(mem_list.head -> allocated);
      if (error != DDI_SUCCESS)
	{
	  printk("klist_empty() : uh-oh, this is not good. :(\n");
	  return (DDI_FAILURE);
	}
    }
  return (DDI_SUCCESS);
}


/***********************************/
/*kfree() : this is the linux call */
/*to free kernel memory            */
/***********************************/
void kfree(const void *ptr) 
{ 
  klist_remove((void *)ptr);
  //DEBUGCMD(printk("kfree() : total memory is now : %ld\n",getTotalMemAllocated()));
} 
/**************************************/
/*kmalloc() : this is the linux call  */
/*to allocate kernel memory           */
/**************************************/
void *kmalloc(size_t size,int flag) 
{ 
  void *ans = NULL;
  //DEBUGCMD(printk("kmalloc() : allocating %d memory\n",size));
  ans = klist_add(size,flag);
  
  //printk("kmalloc() : allocated %d, total allocated %ld\n",size,getTotalMemAllocated());
  return ans;
} 

/**************************************/
/*kmalloc_normal() : normal allocaton */
/* of memory is performed.            */
/**************************************/
kmemptr_t *kmalloc_normal(size_t size, int flag)
{

  kmemptr_t *ptr = NULL;
  ptr = kmem_alloc(sizeof(kmemptr_t),KM_NOSLEEP);
  if (ptr == NULL)
    {
      printk("kmalloc_normal() : cannot allocate kmemptr_t!\n");
      printAlloc();
      goto error;
    }
  ptr -> allocated = kmem_alloc(size,KM_NOSLEEP);

  if (ptr -> allocated == NULL)
    {
      printk("kmalloc_normal() : cannot allocate memory!\n");
      printAlloc();
      goto error_kmemptr;
    }

  ptr -> size = size;
  ptr -> flags = flag;
  ptr -> next = NULL;
  
  return ptr;

 error_kmemptr:
  kmem_free(ptr,sizeof(kmemptr_t));
 error:
  return NULL;

}

int fini_dma_handle(ddi_dma_handle_t *ptr)
{
  ddi_dma_free_handle(ptr);
  return 0;
}

int init_dma_handle(ddi_dma_handle_t *ptr)
{
  int ret;
  
  bzero(ptr,sizeof(ddi_dma_handle_t));

  ret = ddi_dma_alloc_handle(mem_list.dip,
			     &(mem_list.dma_attr),
			     DDI_DMA_DONTWAIT,
			     NULL,
			     /*
			     DDI_DMA_SLEEP,
			     NULL,
			     */
			     ptr);

  if (ret != DDI_SUCCESS)
    {
      pits_dma_alloc_handle_error(ret);
    }
  return ret;

}




kmemptr_t *get_kmemptr(void *ptr)
{
  kmemptr_t *nextptr = NULL;
  kmemptr_t *prevptr = NULL;
  int found = FALSE;
  //grab the mutex.
  mutex_enter(&(mem_list.mutex));

  //scan through the list, trying to find the allocated pointer.
  nextptr = mem_list.head;
  
  while (nextptr != NULL && found == FALSE)
    {
      if (ptr >= nextptr -> allocated && 
	  ptr <= (nextptr -> allocated + nextptr -> size))
	{
	  found = TRUE;
	} 
      else
	{
	  prevptr = nextptr;
	  nextptr = nextptr -> next;
	}
    }
  mutex_exit(&(mem_list.mutex));
  /*didn't find the memory*/
  if (found == FALSE)
    {
      printk("couldn't find the memory in the list! %lu\n",(caddr_t)ptr);
      printAlloc();
      return NULL;
    }
  return nextptr;
}



static int pits_dma_alloc_mem_cb(caddr_t arg)
{
  pits_kernel_dma_memory_t *dma = (pits_kernel_dma_memory_t*)arg;
  int status;

  printk("pits_dma_alloc_mem_cb():start\n");

  //mutex_enter(&mem_list.dma_mutex);
  status = ddi_dma_mem_alloc(dma -> handle,
			     dma -> req_size,
			     &(mem_list.dev_acc_attr),
			     dma -> type,
			     DDI_DMA_DONTWAIT,
			     NULL,
			     &(dma -> addr),
			     &(dma -> real_length),
			     &(dma -> mem_handle));

  // mutex_exit(&(mem_list.dma_mutex));
  if (status != DDI_SUCCESS)
    {

      printk("pits_dma_alloc_mem_cb: error allocating memory, return runout...\n");
      return DDI_DMA_CALLBACK_RUNOUT;
    }
  printk("pits_dma_alloc_mem_cb: returning callback done\n");
  return DDI_DMA_CALLBACK_DONE;

}

/****************************************/
/*kmalloc_dma() : allocates memory that */
/*can be used with dma without scatter  */
/*gathering.                            */
/****************************************/
kmemptr_t *kmalloc_dma(size_t size,int flag)
{
  kmemptr_t *ptr = NULL;
  //pits_kernel_state_t *pk_state = NULL; 
  int status;
  /*
  pk_state = getPitsKernelState();
  if (pk_state == NULL)
    {
      printk("kmalloc_dma() : error getting kernel state!\n");
      goto error;
    }
  */
  //printk("kmalloc_dma()\n");
  ptr = kmem_alloc(sizeof(kmemptr_t),KM_NOSLEEP);
  if (ptr == NULL)
    {
      printk("kmalloc_dma() : error allocating kmemptr!\n");
      printAlloc();
      goto error;
    }

  ptr -> dma = kmem_zalloc(sizeof(pits_kernel_dma_memory_t),KM_NOSLEEP);
  if (ptr -> dma == NULL)
    {
      printk("kmalloc_dma() : error allocating pits_kernel_memory struct!\n");
      printAlloc();
      goto error_kmemptr;
    }

  //printk("init dma handle()\n");

  ptr -> dma -> handle = alloc_dma_handle(KM_SLEEP);
  /*
  ptr -> dma -> handle = kmem_alloc(sizeof(ddi_dma_handle_t),KM_SLEEP);
  status = init_dma_handle(&(ptr -> dma -> handle));
  */


  if (ptr -> dma -> handle == NULL)
    {
      printk("kmalloc_dma() : error allocating dma_handle!\n");
      goto error_dma;
    }
  
  ptr -> dma -> req_size = size;
  ptr -> dma -> type = DDI_DMA_STREAMING;
  //printk("calling dma_alloc\n");
  status = ddi_dma_mem_alloc(*(ptr -> dma -> handle),
			     ptr -> dma -> req_size,
			     &(mem_list.dev_acc_attr),
			     ptr -> dma -> type,
			     /*
			     pits_dma_alloc_mem_cb,
			     (caddr_t)ptr -> dma,
			     */
			     DDI_DMA_DONTWAIT,
			     NULL,
			     &(ptr -> dma -> addr),
			     &(ptr -> dma -> real_length),
			     &(ptr -> dma -> mem_handle));


  //printk("done calling dma_alloc\n");
  if (status != DDI_SUCCESS)
    {
      pits_dma_alloc_mem_error(status);
      goto error_handle;
    }

  if (size != ptr -> dma -> real_length)
    {
      printk("kmalloc_dma() : size %d <> real_length %d\n",size, ptr -> dma -> real_length);
      printk("this may be where the bug is....\n");
    }

  ptr -> allocated = (void*)ptr -> dma -> addr;
  ptr -> size = size;
  ptr -> flags = flag;
  ptr -> next = NULL;
  //printk("done kmalloc_dma()\n");
  return ptr;

 error_handle:
  /*
  ddi_dma_free_handle(&(ptr -> dma -> handle));
  kmem_free(&(ptr -> dma -> handle,sizeof(ddi_dma_handle_t));
  */
  release_dma_handle(ptr -> dma -> handle);
error_dma:
  kmem_free(ptr -> dma,sizeof(pits_kernel_dma_memory_t));
 error_kmemptr:
  kmem_free(ptr,sizeof(kmemptr_t));
 error:
  return NULL;
}


void kfree_dma(kmemptr_t *ptr)
{
  ddi_dma_mem_free(&(ptr -> dma -> mem_handle));
  //ddi_dma_free_handle(ptr -> dma -> handle);
  //kmem_free(ptr -> dma -> handle);
  release_dma_handle(ptr -> dma -> handle);
  kmem_free(ptr -> dma,sizeof(pits_kernel_dma_memory_t));
  kmem_free(ptr,sizeof(kmemptr_t));
}

void kfree_normal(kmemptr_t *ptr)
{
  kmem_free(ptr -> allocated,ptr -> size);
  kmem_free(ptr,sizeof(kmemptr_t));
}
/***********************************************/
/*pits_dma_alloc_handle_error()                */
/***********************************************/
void pits_dma_alloc_handle_error(int status)
{
  printk("error allocating dma handle: %d -",status);
  switch (status)
    {
    case DDI_DMA_BADATTR:
      printk("DDI_DMA_BADATTR\n");
      break;
    case DDI_DMA_NORESOURCES:
      printk("DDI_DMA_NORESOURCES\n");
      break;
    default:
      printk("unknown error\n");
    }

}

/**********************************************/
/*pits_dma_alloc_mem_error()                  */
/**********************************************/   
void pits_dma_alloc_mem_error(int status)
{
  printk("error allocating dma memory: %d - failure\n",status);
}

/******************************************/
/*pits_dma_bind_handle_error()            */
/******************************************/
void pits_dma_bind_handle_error(int status)
{
  printk("error binding dma handle: %d -",status);
  switch(status)
    {
    case DDI_DMA_PARTIAL_MAP:
      printk("DDI_DMA_PARTIAL_MAP\n");
      break;
    case DDI_DMA_INUSE:
      printk("DDI_DMA_INUSE\n");
      break;
    case DDI_DMA_NORESOURCES:
      printk("DDI_DMA_NO_RESOURCES\n");
      break;
    case DDI_DMA_NOMAPPING:
      printk("DDI_DMA_NOMAPPING\n");
      break;
    case DDI_DMA_TOOBIG:
      printk("DDI_DMA_TOOBIG\n");
      break;
    default:
      printk("unknown error\n");
    };
}


void printAlloc()
{
  printk("Total Allocated- %lu\n",getTotalMemAllocated());
  printk("Total Allocations - %lu\n",getTotalAllocations());
}





/***************************************/
/*getTotalMemAllocated():              */
/*This function will return the total  */
/*amount of memory allocated in bytes. */
/***************************************/
uint64_t getTotalMemAllocated()
{
  uint64_t total = 0;
  mutex_enter(&(mem_list.mutex));
  total = mem_list.total_allocated;
  mutex_exit(&(mem_list.mutex));
  return total;
}
/*********************************************/
/*getTotalAllocations(): this function will  */
/*get the current total allocations.         */
/*********************************************/
uint64_t getTotalAllocations()
{
  uint64_t total = 0;
  mutex_enter(&(mem_list.mutex));
  total = mem_list.total_allocations;
  mutex_exit(&(mem_list.mutex));
  return total;
}
